/*
 * Sidereal: a sidereal clock for PalmOS.
 *
 * Copyright (C) 2001  Edgar Bonet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This calculation is based on the definition of universal time given
 * by Aoki et al. [1], quoted here from Rick Fisher [2]:
 *
 * "GMST (in seconds at UT1=0) = 24110.54841 + 8640184.812866 * T
 *                                 + 0.093104 * T^2 - 0.0000062 * T^3
 * where T is in Julian centuries from 2000 Jan. 1 12h UT1"
 *
 * Linearizing around T = 0.2 (2020-01-01, 12:00 UT1) gives:
 *
 *      GMST(UT1=0) = 24147.5147074096 + 8640184.850108344 * (T-0.2)
 *
 * which is accurate to within 3.8 ms in the interval [2000, 2040]. From
 * this linearized equation we can get GMST at the epoch that we define
 * to be 1904-01-01, 00:00 UT1 (instead of local time as in PalmOS):
 *
 *      T(epoch) = - (96 * 365.25 + 0.5) / 36525 = -.9600136892539357
 *      GMST(epoch) = 23814.8108972664
 *
 * as well as the number of sidereals seconds per second:
 *
 *      u = (8640184.850108344 + 3155760000) / 3155760000
 *        = 1.00273790936259675767
 *
 * (3155760000 is the number of seconds in a julian century).
 * Then, if t is the number of seconds since the epoch, GMST is just
 *
 *      GMST = GMST(epoch) + u * t.
 *
 * In order to calculate u * t using 32 bit integers without
 * overflowing, u and t are broken down in 16 bit chunks:
 *
 *      u = 1 + c*2^(-24) + d*2^(-40)
 *      t = x*2^16 + y
 *      GMST = GMST(epoch) + t + c*x*2^(-8)
 *           + (d*x + c*y)*2^(-24) + d*y*2^(-40)
 *
 * wich is calculated in units of 2^(-8) s (the last term is less than
 * 2^(-8) s, so we drop it). In principle we could give the sidereal
 * time with sub-second accuracy if we knew the fractional part of t as
 * well as UT1-UTC.
 *
 * References:
 *
 *  [1] S. Aoki, B. Guinot, G. H. Kaplan, H. Kinoshita, D. D. McCarthy,
 *      and P. K. Seidelmann, Astron. Astrophys. 105, 359-361 (1982).
 *  [2] <http://sadira.gb.nrao.edu/~rfisher/Ephemerides/times.html#GMST>
 */

#include <PalmOS.h>

extern Int32 time_zone;     /* in seconds east of Greenwich */
extern UInt32 longitude;    /* in 1/256 s east of Greenwich */

UInt32 local_to_lmst(UInt32 t)
{
	const UInt32 a = 6096592, b = 86400, c = 45934, d = 32556;
	UInt32 x, y, st;

	t -= time_zone;
	x = t >> 16;
	y = t & 0xffff;
	st = a + longitude + (t%b<<8) + c*x + (d*x>>16) + (c*y>>16);
	st %= b << 8;
	return (st+128) >> 8;	/* Adding 128 zeroes the average error. */
}

/* vim: set ts=4 sw=4: */
