/*
 * Sidereal: a sidereal clock for PalmOS.
 *
 * Copyright (C) 2001  Edgar Bonet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include "sidereal.h"

UInt32 local_to_lmst(UInt32 t);

/* Timezone and longitude as displayed in the Location form and saved in
 * the preferences. */
typedef struct {
	Int32 tz;	/* minutes east of UTC */
	Int32 lg;	/* arc seconds east of Greenwich */
} location_t;


/***********************************************************************
 * Globals
 */


/* Current location. */
static location_t location;

/* Default: EDT = UTC - 4h, Fuertes observatory: -76 28' 29" */
static location_t default_location = {-240, -275309};

/* This is our internal representation of the current location. */
Int32 time_zone;	/* in s */
UInt32 longitude;	/* in 1/256 s */

/* Interface colors. */
static Boolean use_color = false;
typedef enum {normal, night} color_mode_t;
static RGBColorType text_rgb __attribute__ ((aligned)) = {0, 0x99, 0x33 ,0};
static IndexedColorType text_color;


/***********************************************************************
 * Drawing functions
 */

/* Height difference between the font used for times and the font used
 * for labels. */
static Int16 font_correction;

/* Preferred format for displaying the date. */
DateFormatType date_format;

static color_mode_t color_mode = normal;

static void draw_time(UInt32 time, Int16 x, Int16 y)
{
	DateTimeType date_time;
	Char s[] = "00:00:00";

	TimSecondsToDateTime(time, &date_time);
	s[0] = '0' + date_time.hour / 10;
	s[1] = '0' + date_time.hour % 10;
	s[3] = '0' + date_time.minute / 10;
	s[4] = '0' + date_time.minute % 10;
	s[6] = '0' + date_time.second / 10;
	s[7] = '0' + date_time.second % 10;
	WinDrawChars(s, StrLen(s), x, y);
}

static void draw_date(UInt32 time, Int16 x, Int16 y)
{
	DateType date;
	Char buffer[dateStringLength];

	DateSecondsToDate(time, &date);
	DateToAscii(date.month, date.day, 1904+date.year,
			date_format, buffer);
	WinDrawChars(buffer, StrLen(buffer), x, y);
}

static void draw_utc_date(UInt32 time, Int16 x, Int16 y)
{
	draw_date(time - time_zone, x, y);
}

static void draw_utc(UInt32 time, Int16 x, Int16 y)
{
	draw_time(time - time_zone, x, y);
}

static void draw_lmt(UInt32 time, Int16 x, Int16 y)
{
	draw_time(time - time_zone + (longitude>>8), x, y);
}

static void draw_lmst(UInt32 time, Int16 x, Int16 y)
{
	draw_time(local_to_lmst(time), x, y);
}

static void draw_mjd(UInt32 time, Int16 x, Int16 y)
{
	static char buff[] = "00000.00000";
	Int32 days, frac;
	Int16 i;

	time -= time_zone;
	days = 16480 + time / 86400;
	frac = (time % 86400) * 125/108;
	for (i= 4; i>=0; i--, days/=10) buff[i] = '0' + days % 10;
	for (i=10; i>=6; i--, frac/=10) buff[i] = '0' + frac % 10;
	WinDrawChars(buff, StrLen(buff), x, y);
}

/*
 * Equation of time in 1/32 s.
 * Tabulated every 2 days from 2004 jan 02, 0 TT.
 * Data from <http://lep694.gsfc.nasa.gov/code693/TYPE/sun2.html>.
 */
static const Int16 eot_table[] = {
     -6762,  -8547, -10283, -11958, -13573, -15119, -16592, -17986,
    -19297, -20521, -21651, -22684, -23615, -24438, -25159, -25774,
    -26283, -26685, -26984, -27180, -27276, -27275, -27178, -26992,
    -26717, -26355, -25908, -25380, -24775, -24096, -23348, -22533,
    -21662, -20738, -19766, -18755, -17708, -16632, -15530, -14408,
    -13269, -12115, -10957,  -9798,  -8640,  -7489,  -6351,  -5230,
     -4132,  -3064,  -2028,  -1035,    -85,    819,   1672,   2472,
      3216,   3902,   4528,   5095,   5595,   6028,   6391,   6682,
      6899,   7041,   7108,   7099,   7017,   6864,   6639,   6345,
      5988,   5569,   5094,   4564,   3984,   3355,   2685,   1971,
      1222,    441,   -364,  -1189,  -2027,  -2869,  -3709,  -4537,
     -5353,  -6147,  -6914,  -7650,  -8349,  -9008,  -9623, -10188,
    -10702, -11156, -11551, -11881, -12141, -12328, -12439, -12473,
    -12428, -12305, -12099, -11819, -11461, -11030, -10526,  -9950,
     -9303,  -8588,  -7804,  -6951,  -6037,  -5063,  -4031,  -2945,
     -1809,   -627,    595,   1850,   3138,   4446,   5775,   7119,
      8475,   9837,  11203,  12566,  13923,  15268,  16600,  17907,
     19184,  20426,  21627,  22780,  23882,  24927,  25911,  26832,
     27680,  28455,  29153,  29771,  30305,  30751,  31104,  31361,
     31519,  31571,  31516,  31354,  31082,  30700,  30208,  29606,
     28899,  28087,  27176,  26162,  25050,  23843,  22542,  21154,
     19681,  18131,  16507,  14820,  13072,  11273,   9434,   7564,
      5675,   3772,   1864,    -40,  -1930,  -3804,  -5650,  -7462
};

static Int32 eot(UInt32 time)
{
	UInt32 pos;
	Int32 offset;
	Int32 result;

	time -= 153736;    /* epoch = 2004-01-02 0 TT (TT = UTC + 64 s) */
	time %= 31556926;  /* tropical year */
	pos = time / 172800;
	offset = time % 172800;
	result = eot_table[pos];
	result += offset/4*(eot_table[pos+1]-eot_table[pos])/43200;
	return (result + 16) / 32;
}

static void draw_lt(UInt32 time, Int16 x, Int16 y)
{
	static Int32 eot_offset = -9999;  /* invalid */

	if (eot_offset == -9999) eot_offset = eot(time - time_zone);
	draw_time(time - time_zone + (longitude>>8) + eot_offset, x, y);
}

static void (*draw[])(UInt32, Int16, Int16) = {
	draw_time, draw_lt, draw_lmt, draw_lmst,
	draw_utc, draw_date, draw_utc_date, draw_mjd
};

static UInt8 selection[] = {0, 4, 3, 7};

/*
 * Draw all the times and the MJD. redraw forces a redraw even if the
 * current time has not changed since the last call of draw_times().
 */
static void draw_times(Boolean redraw)
{
	static UInt32 old_time = 0;
	UInt32 time;

	time = TimGetSeconds();
	if (redraw) {
		RectangleType rec;

		rec.extent.x = 62;
		rec.extent.y = 10;
		rec.topLeft.x = 73;
		rec.topLeft.y = 37;
		WinEraseRectangle(&rec, 0);
		rec.topLeft.y = 67;
		WinEraseRectangle(&rec, 0);
		rec.topLeft.y = 97;
		WinEraseRectangle(&rec, 0);
		rec.topLeft.y = 127;
		WinEraseRectangle(&rec, 0);
	}
	if (redraw || time != old_time) {
		if (use_color) WinSetTextColor(text_color);
		draw[selection[0]](time, 73,  38 - font_correction);
		draw[selection[1]](time, 73,  68 - font_correction);
		draw[selection[2]](time, 73,  98 - font_correction);
		draw[selection[3]](time, 73, 128 - font_correction);
		old_time = time;
	}
}

/*
 * How the hell was I supposed to know that RGBColorType has to be
 * aligned? This structure only contains bytes!
 */
static void swap_colors(void)
{
	UIColorTableEntries i;
	static RGBColorType defaults[UILastColorTableEntry]
		__attribute__ ((aligned));
	static IndexedColorType	default_text;
	static RGBColorType red __attribute__ ((aligned)) = {0, 255, 0 ,0};
	static RGBColorType	black __attribute__ ((aligned)) = {0, 0, 0, 0};
	static Boolean defaults_saved = false;
	FormPtr form;

	if (!defaults_saved) {
		for (i=0; i<UILastColorTableEntry; i++)
			UIColorGetTableEntryRGB(i, &defaults[i]);
		default_text = text_color;
		defaults_saved = true;
	}
	if (color_mode == normal) {
		for (i=0; i<UILastColorTableEntry; i++) {
			RGBColorType rgb = defaults[i];
			/* grey = .299 red + .587 green + .114 blue */
			if (38*rgb.r + 75*rgb.g + 15*rgb.b > 19584)
				UIColorSetTableEntry(i, &black);
			else
				UIColorSetTableEntry(i, &red);
		}
		text_color = WinRGBToIndex(&red);
		color_mode = night;
	} else {
		for (i=0; i<UILastColorTableEntry; i++) {
			UIColorSetTableEntry(i, &defaults[i]);
		}
		text_color = default_text;
		color_mode = normal;
	}
	form = FrmGetActiveForm();
	/* The new colors will be ignored by WinEraseWindow() unless we
	 * redraw the form fist. */
	if (form) FrmDrawForm(form);
	WinEraseWindow();
	if (form) FrmDrawForm(form);
	draw_times(true);
}

/***********************************************************************
 * Location form
 */

/*
 * This is a destructive read! To be used only when we are done with the
 * form.
 */
static Char *get_field_text(FormPtr form, UInt16 id, Char *buffer, int length)
{
	FieldPtr field;
	MemHandle handle;
	Char *text;

	field = FrmGetObjectPtr(form, FrmGetObjectIndex(form, id));
	handle = FldGetTextHandle(field);
	FldSetTextHandle(field, NULL);
	text = MemHandleLock(handle);
	StrNCopy(buffer, text, length-1);
	buffer[length-1] = chrNull;
	MemHandleFree(handle);
	return buffer;
}

static void set_field_text(FormPtr form, UInt16 id, Char *text, int length)
{
	FieldPtr field;
	MemHandle old_handle, handle;

	if (!length) length = StrLen(text) + 1;
	field = FrmGetObjectPtr(form, FrmGetObjectIndex(form, id));
	old_handle = FldGetTextHandle(field);
	handle = MemHandleNew(length);
	StrCopy(MemHandleLock(handle), text);
	MemHandleUnlock(handle);
	FldSetTextHandle(field, handle);
	if (old_handle) MemHandleFree(old_handle);
}

static Int16 get_control_value(FormPtr form, UInt16 id)
{
	return FrmGetControlValue(form, FrmGetObjectIndex(form, id));
}

static void set_control_value(FormPtr form, UInt16 id, Int16 value)
{
	FrmSetControlValue(form, FrmGetObjectIndex(form, id), value);
}

/* Convert location_t to our internal representation. */
static void update_location(void)
{
	Int32 long_arcs;

	time_zone = 60 * location.tz;
	long_arcs = location.lg;
	while (long_arcs < 0) long_arcs += 1296000;
	longitude = long_arcs * 256 / 15;
}

static void change_location(void)
{
	FormPtr previous_form, form;
	Int16 button;
	Int32 tz_h, tz_min, long_deg, long_min, long_s;
	Char buff[4];

	previous_form = FrmGetActiveForm();
	form = FrmInitForm(location_form);
	FrmSetActiveForm(form);

	/* Initialize the fields. */
	tz_min = location.tz;
	if (tz_min < 0) {
		tz_min = -tz_min;
		set_control_value(form, location_tz_w, 1);
	} else
		set_control_value(form, location_tz_e, 1);
	tz_h = tz_min / 60; tz_min %= 60;
	set_field_text(form, location_tz_h, StrIToA(buff, tz_h), sizeof buff);
	set_field_text(form, location_tz_min, StrIToA(buff, tz_min), sizeof buff);
	long_s = location.lg;
	if (long_s < 0) {
		long_s = -long_s;
		set_control_value(form, location_long_w, 1);
	} else
		set_control_value(form, location_long_e, 1);
	long_min = long_s / 60; long_s %= 60;
	long_deg = long_min / 60; long_min %= 60;
	set_field_text(form, location_long_deg,
			StrIToA(buff, long_deg), sizeof buff);
	set_field_text(form, location_long_min,
			StrIToA(buff, long_min), sizeof buff);
	set_field_text(form, location_long_s,
			StrIToA(buff, long_s), sizeof buff);

	/* Run the dialog. */
	button = FrmDoDialog(form);

	/* Cancel? */
	if (button == location_cancel) goto cleanup;

	/* Update our location. */
	get_field_text(form, location_tz_h, buff, sizeof buff);
	tz_h = StrAToI(buff);
	get_field_text(form, location_tz_min, buff, sizeof buff);
	tz_min = StrAToI(buff);
	location.tz = 60*tz_h + tz_min;
	if (get_control_value(form, location_tz_w))
		location.tz = -location.tz;
	get_field_text(form, location_long_deg, buff, sizeof buff);
	long_deg = StrAToI(buff);
	get_field_text(form, location_long_min, buff, sizeof buff);
	long_min = StrAToI(buff);
	get_field_text(form, location_long_s, buff, sizeof buff);
	long_s = StrAToI(buff);
	location.lg = 3600*long_deg + 60*long_min + long_s;
	if (get_control_value(form, location_long_w))
		location.lg = -location.lg;
	update_location();

cleanup:
	FrmSetActiveForm(previous_form);
	FrmDeleteForm(form);
}


/***********************************************************************
 * Main form
 */

static Boolean main_form_handler(EventPtr event)
{
	static FormPtr form;

	switch (event->eType) {
		case frmOpenEvent:
			form = FrmGetActiveForm();
			FrmDrawForm(form);
			draw_times(true);
			break;
		case frmCloseEvent:
			FrmEraseForm(form);
			FrmDeleteForm(form);
			form = 0;
			break;
		case popSelectEvent:
			selection[event->data.popSelect.listID - main_list_0] =
				event->data.popSelect.selection;
			draw_times(true);
			return false;
		default:
			return false;
	}
	return true;
}


/***********************************************************************
 * Main and friends
 */

static UInt32 romVersion;
static FormPtr main_form_ptr = NULL;
static struct {
	color_mode_t color_mode;
	UInt8 selection[4];
} unsaved_prefs;

static void start_application(void)
{
	UInt16 location_size = sizeof(location);
	UInt16 unsaved_prefs_size = sizeof(unsaved_prefs);
	Int16 version, i;

	/* Select the biggest font available with current ROM. */
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion >= 0x03003000) {
		FntSetFont(largeBoldFont);
		font_correction = 3;
	} else {
		FntSetFont(boldFont);
		font_correction = 0;
	}

	/* Get the preferred date format. */
	date_format = PrefGetPreference(prefDateFormat);

	/* Get the location from the saved preferences. */
	version = PrefGetAppPreferences(APID, 1,
			&location, &location_size, true);
	if (version == noPreferenceFound) location = default_location;
	update_location();

	/* Get the color and selections from the unsaved preferences. */
	version = PrefGetAppPreferences(APID, 1,
				&unsaved_prefs, &unsaved_prefs_size, false);
	if (version == noPreferenceFound) color_mode = normal;
	else {
		color_mode = unsaved_prefs.color_mode;
		if (unsaved_prefs_size == sizeof(unsaved_prefs))
			for (i=0; i<4; i++) selection[i] = unsaved_prefs.selection[i];
	}

	/* Do we use colors? */
	if (romVersion >= 0x03503000) {
		UInt32 screen_depth;

		WinScreenMode(winScreenModeGet, NULL, NULL, &screen_depth, NULL);
		use_color = screen_depth >= 8;
	} else use_color = false;

	if (use_color) text_color = WinRGBToIndex(&text_rgb);

	if (use_color && color_mode == night) {
		color_mode = normal;
		swap_colors();
	}

	FrmGotoForm(main_form);
}

static void set_selection(FormPtr form, int list_id,
		int popup_id, int value)
{
	ListType *list;
	ControlPtr popup;

	list = FrmGetObjectPtr(form, FrmGetObjectIndex(form, list_id));
	popup = FrmGetObjectPtr(form, FrmGetObjectIndex(form, popup_id));
	LstSetSelection(list, value);
	CtlSetLabel(popup, LstGetSelectionText(list, value));
}

static Boolean application_handler(EventPtr event)
{
	static Boolean active = false;

	switch (event->eType) {
		case frmLoadEvent: {
			FormPtr pForm = FrmInitForm(event->data.frmLoad.formID);
			FrmSetActiveForm(pForm);
			if (event->data.frmLoad.formID == main_form) {
				FrmSetEventHandler(pForm, main_form_handler);
				set_selection(pForm, main_list_0,
						main_trigger_0, selection[0]);
				set_selection(pForm, main_list_1,
						main_trigger_1, selection[1]);
				set_selection(pForm, main_list_2,
						main_trigger_2, selection[2]);
				set_selection(pForm, main_list_3,
						main_trigger_3, selection[3]);
				main_form_ptr = pForm;
			}
			break;
		}
		case menuOpenEvent:
			if (use_color)
				MenuAddItem(menu_location, menu_color, '\0', "Change colors");
			break;
		case winEnterEvent:
			if (event->data.winEnter.enterWindow == (WinHandle) main_form_ptr)
				active = true;
			break;
		case winExitEvent:
			if (event->data.winExit.exitWindow == (WinHandle) main_form_ptr)
				active = false;
			break;
		case menuEvent:
			switch (event->data.menu.itemID) {
				case menu_help:
					FrmHelp(help_message);
					break;
				case menu_about:
					FrmAlert(about_alert);
					break;
				case menu_location:
					change_location();
					break;
				case menu_color:
					swap_colors();
					break;
			}
			break;
		default:
			if (active) draw_times(false);
			return false;
	}
	return true;
}

static void stop_application(void)
{
	Int16 i;

	PrefSetAppPreferences(APID, 1, 1, &location, sizeof location, true);
	unsaved_prefs.color_mode = color_mode;
	for (i=0; i<4; i++) unsaved_prefs.selection[i] = selection[i];
	PrefSetAppPreferences(APID, 1, 1,
				&unsaved_prefs, sizeof unsaved_prefs, false);
	FrmCloseAllForms();
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	EventType event;
	UInt16 error;

	(void) cmdPBP; (void) launchFlags;

	if (cmd != sysAppLaunchCmdNormalLaunch) return 0;

	start_application();
	do {
		EvtGetEvent(&event, SysTicksPerSecond()/4);
		if (SysHandleEvent(&event)) continue;
		if (MenuHandleEvent(0, &event, &error)) continue;
		if (application_handler(&event)) continue;
		FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
	stop_application();
	return 0;
}

/* vim: set ts=4 sw=4: */
